/*
  simplezdb.cpp - a ruby extension module providing the simple access to ZDB
  $Id: simplezdb.cpp,v 1.6 2005/11/21 01:18:58 noir Exp $
  Copyright (C) 2005 Hiroyuki KUROSAKI <noir (at) st . rim . or . jp>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ruby.h"

#include <sl/slzdb.h>
#include <qdatetime.h>

static VALUE mSimpleZDB;
static VALUE cDataBase;
static VALUE cItemAttr;
static VALUE cCard;
static VALUE cCategorySet;

VALUE cDate, cDateTime;

namespace DataBase {
    enum DataBaseType {
        ADDRESSBOOK = 0,
        TODOLIST,
        DATEBOOK,
        MAIL_INBOX,
        MAIL_OUTBOX,
        MEMO,
    };
}

static const char* DATABASE_TYPES[] = {
    "ADDRESSBOOK",
    "TODOLIST",
    "DATEBOOK",
    "MAIL_INBOX",
    "MAIL_OUTBOX",
    "MEMO",
};

/*
 - It doesn't work; parameters are different between them.

static const QString (*db_file_name_funcs[])(bool) = {
    (const QString (*)(bool))SlZDataBase::addressbookFileName,
    (const QString (*)(bool))SlZDataBase::todolistFileName,
    (const QString (*)(bool))SlZDataBase::datebookFileName,
    (const QString (*)(bool))SlZDataBase::mailInboxFileName,
    (const QString (*)(bool))SlZDataBase::mailOutboxFileName,
    (const QString (*)(bool))SlZDataBase::memoFileName,
};

 - It works, but we cannot get memoItems via function.
   Because we cannot find the declaration of struct SlZdbItemInfo.

static SlZDataBaseItemInfo (*db_item_info_funcs[])() = {
    SlZDataBase::addressbookItems,
    SlZDataBase::todolistItems,
    SlZDataBase::datebookItems,
    SlZDataBase::maillistItems,
    SlZDataBase::maillistItems,
};

*/

static void zdb_free(void* p)
{
    SlZDataBase* p_zdb = static_cast<SlZDataBase*>(p);

    if (p_zdb->isOpened())
        p_zdb->close();

    delete p_zdb;
}


SlZDataBase* zdb_open(VALUE v_db_type, VALUE v_is_readonly, VALUE v_filename)
{
    int   db_type      = NUM2INT(v_db_type);
    bool  is_readonly  = RTEST(v_is_readonly);

    QString filename(RTEST(v_filename) ? StringValuePtr(v_filename) : "");

    SlZDataBase *p_zdb;

/*
 - Abandoned for the reasons described above.

    if (db_type < (sizeof(DATABASE_TYPES)/sizeof(char*))) {
	p_zdb = new SlZDataBase((*(db_file_name_funcs[db_type]))(false),
				(*(db_item_info_funcs[db_type]))(),
				0,
				is_readonly);
    }
    else {
        rb_raise(rb_eRuntimeError, "No such database type: %d", db_type);
    }
*/

    switch (db_type) {
    case DataBase::ADDRESSBOOK:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::addressbookFileName(),
                                SlZDataBase::addressbookItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::TODOLIST:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::todolistFileName(),
                                SlZDataBase::todolistItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::DATEBOOK:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::datebookFileName(),
                                SlZDataBase::datebookItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::MAIL_INBOX:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::mailInboxFileName(),
                                SlZDataBase::maillistItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::MAIL_OUTBOX:
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::mailOutboxFileName(),
                                SlZDataBase::maillistItems(),
                                0,
                                is_readonly);
        break;
    case DataBase::MEMO:

/* thanks to ZdbManager.cpp in zdbat-1.0.1 written by yakty */
    {
        SlZDataBaseItemInfo items;

	items.addItemInfo(1, "CLAS", SlZDataBaseItemInfo::TEXT, "title");
	items.addItemInfo(2, "MEM1", SlZDataBaseItemInfo::TEXT, "contents");
	items.addItemInfo(3, "DATE", SlZDataBaseItemInfo::TIME, "modified time");
        p_zdb = new SlZDataBase(filename.length() > 0
                                  ? filename
                                  : SlZDataBase::memoFileName(),
                                items,
                                0,
                                is_readonly);
    }
/* end */
        break;
    default:
        rb_raise(rb_eRuntimeError, "No such database type: %d", db_type);
    }

    return p_zdb;
}

SlZDataBase* db_get_zdb(VALUE self)
{
    SlZDataBase* p_zdb;
    VALUE zdb = rb_iv_get(self, "zdb");

    if (NIL_P(zdb)) {
        rb_raise(rb_eRuntimeError, "Database not initialized.");
        return 0;
    }

    Data_Get_Struct(zdb, SlZDataBase, p_zdb);

    return p_zdb;
}

static VALUE db_close(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return p_zdb->close() ? Qtrue : Qfalse;
}

VALUE db_init_item_attrs(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    VALUE item_attrs = rb_ary_new();
    rb_iv_set(self, "@item_attrs", item_attrs);

    size_t n_items(p_zdb->items().count());

    for (size_t i = 0; i < n_items; ++i) {
        QString name(p_zdb->itemName(i).left(4));

        if (!p_zdb->nameToFieldId(i))
            name = name.sprintf("NONAME%d", i);

        uint value_type(QChar(p_zdb->itemName(i).at(4)).latin1());

	VALUE item_attr = rb_funcall(cItemAttr, rb_intern("new"), 2,
                                     rb_str_new2((const char*)name),
                                     INT2FIX(value_type));
	
        rb_ary_push(item_attrs, item_attr);
    }
    return self;
}

/*
 * Document-method: initialize
 * call-seq:
 *   DataBase.new(database_type, is_readonly = false, filename = nil) -> database
 * Open the database and return the DataBase object.
 * Possible values of database_type are:
 *
 *   DataBase::ADDRESSBOOK
 *   DataBase::TODOLIST
 *   DataBase::DATEBOOK
 *   DataBase::MAIL_INBOX
 *   DataBase::MAIL_OUTBOX
 *   DataBase::MEMO
 */

static VALUE db_initialize(VALUE self, VALUE args)
{
    VALUE db_type = rb_funcall(args, rb_intern("[]"), 1, INT2FIX(0));

    if (NIL_P(db_type)) {
        rb_warn("Database type not specified.");
        return self;
    }
    
    VALUE is_readonly = rb_funcall(args, rb_intern("[]"), 1, INT2FIX(1));
    VALUE filename    = rb_funcall(args, rb_intern("[]"), 1, INT2FIX(2));
    
    SlZDataBase* p_zdb = zdb_open(db_type, is_readonly, filename);

    if (!p_zdb) {
        rb_raise(rb_eRuntimeError, "Open database failed.");
        return self;
    }

    VALUE zdb = Data_Wrap_Struct(self, 0, zdb_free, p_zdb);

    rb_iv_set(self, "zdb", zdb);

    db_init_item_attrs(self);

    return self;
}

/*
 * Document-method: open
 * call-seq:
 *   DataBase.open(database_type, is_readonly = false, filename = nil) -> database
 *   DataBase.open(database_type, is_readonly = false, filename = nil) {|database| ... }
 *
 * Open the database specified by the parameters.
 * See also DataBase.new.
 * If block given, open database, provide it as the block parameter,
 * execute the block, and close the database.
 */
static VALUE db_s_open(int argc, VALUE* argv, VALUE klass)
{
    VALUE obj = rb_funcall2(klass, rb_intern("new"), argc, argv);
    if (NIL_P(obj))
	return Qnil;
    if (rb_block_given_p()) {
	return rb_ensure(RUBY_METHOD_FUNC(rb_yield),
			 obj,
			 RUBY_METHOD_FUNC(db_close),
			 obj);
    }
    return obj;
}

/*
 * Document-method: size
 * call-seq:
 *   database.size -> fixnum
 */

static VALUE db_size(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return INT2FIX(p_zdb->count());
}

/*
 * Document-method: each
 * call-seq:
 *     database.each {|card| ... } -> database
 */

static VALUE db_each(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    VALUE card;

    for (CardId card_id = p_zdb->firstCard();
         card_id != 0;
         card_id = p_zdb->nextCard(card_id)) {
        card = rb_funcall(cCard,
                          rb_intern("new"),
                          2,
                          self,
                          ULONG2NUM(card_id));
        rb_yield(card);
    }
    return self;
}

VALUE db_get_item_attr_with_index(VALUE self, VALUE key)
{
    VALUE result = Qnil;
    VALUE item_attrs = rb_iv_get(self, "@item_attrs");

    int size = NUM2INT(rb_funcall(item_attrs, rb_intern("size"), 0));

    for (int i = 0; i < size; ++i) {
        VALUE attr = rb_funcall(item_attrs, rb_intern("[]"), 1, INT2FIX(i));
        VALUE name = rb_funcall(attr, rb_intern("name"), 0);
        if (RTEST(rb_funcall(key, rb_intern("=="), 1, name))) {
            result = rb_ary_new();
            rb_ary_push(result, INT2FIX(i));
            rb_ary_push(result, attr);
            return result;
        }
    }
    return result;
}

/*
 * Document-method: read_time_field
 * call-seq:
 *   database.read_time_field(field, card_id) -> datetime or date
 *
 * (Normally not used directly. Called from Card.)
 * Return the date/time value specified by the field (i.e., index of
 * item_attrs) of the card card_id.
 */
static VALUE db_read_time_field(VALUE self, VALUE v_field, VALUE v_card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    int    field   = NUM2INT(v_field);
    CardId card_id = NUM2ULONG(v_card_id);

    QDateTime q_datetime = p_zdb->readTimeField(field, card_id); // UTC
    QDate     q_date     = q_datetime.date();
    QTime     q_time     = q_datetime.time();

    VALUE val = Qnil;

    if (q_datetime.isValid()) {
        val = rb_funcall(cDateTime, rb_intern("new"), 7,
                         INT2FIX(q_date.year()),
                         INT2FIX(q_date.month()),
                         INT2FIX(q_date.day()),
                         INT2FIX(q_time.hour()),
                         INT2FIX(q_time.minute()),
                         INT2FIX(q_time.second()),
                         INT2FIX(0));
    }
    else if (q_date.isValid()) {
        val = rb_funcall(cDate, rb_intern("new"), 3,
                         INT2FIX(q_date.year()),
                         INT2FIX(q_date.month()),
                         INT2FIX(q_date.day()));
    }

    return val;
}

/*
 * Document-method: read_categories
 * call-seq:
 *   database.read_categories(card_id) -> array
 *
 * (Normally not used directly. Called from Card)
 * Return the categories of the card card_id.
 */

static VALUE db_read_categories(VALUE self, VALUE card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    VALUE category_set = rb_iv_get(self, "@category_set");

    if (NIL_P(category_set)) {
        category_set = rb_funcall(cCategorySet, rb_intern("new"), 0);
        rb_iv_set(self, "@category_set", category_set);
    }

    if (RTEST(rb_funcall(card_id, rb_intern("respond_to?"), 1,
                         rb_str_new2("card_id")))) {
        card_id = rb_funcall(card_id, rb_intern("card_id"), 0);
    }

    VALUE val = rb_ary_new();

    QArray<int> ids = p_zdb->readCategories(NUM2ULONG(card_id));
    for (QArray<int>::ConstIterator it = ids.begin(); it != ids.end(); ++it) {
        rb_ary_push(val,
                    rb_funcall(category_set, rb_intern("[]"), 1,
                               INT2FIX((*it))));
    }
    return val;
}

/*
 * Document-method: read_item
 * call-seq:
 *   database.read_item(key, card_id) -> object
 *
 * (Normally not used directly. Called from Card)
 * Return the value specified by key of the card card_id.
 */

static VALUE db_read_item(VALUE self, VALUE key, VALUE v_card_id)
{
    VALUE item_attr_with_index = db_get_item_attr_with_index(self, key);
    if (NIL_P(item_attr_with_index))
        return Qnil;

    VALUE item_attr_index = rb_funcall(item_attr_with_index,
                                       rb_intern("[]"), 1,
                                       INT2FIX(0));

    int field = NUM2INT(item_attr_index);

    VALUE item_attr = rb_funcall(item_attr_with_index,
                                 rb_intern("[]"), 1,
                                 INT2FIX(1));

    int value_type = NUM2INT(rb_funcall(item_attr, rb_intern("value_type"), 0));
    if (RTEST(rb_funcall(v_card_id, rb_intern("respond_to?"),
                         1, rb_str_new2("card_id")))) {
        v_card_id = rb_funcall(v_card_id, rb_intern("card_id"), 0);
    }

    SlZDataBase* p_zdb = db_get_zdb(self);

    CardId card_id = NUM2ULONG(v_card_id);

    VALUE val;

    switch (value_type) {
    case SlZDataBaseItemInfo::TEXT:
        val = rb_str_new2(p_zdb->readField(field, card_id).utf8());
        break;
    case SlZDataBaseItemInfo::UCHAR:
        val = INT2FIX(p_zdb->readUcharField(field, card_id));
        break;
    case SlZDataBaseItemInfo::USHORT:
        val = INT2FIX(p_zdb->readUshortField(field, card_id));
        break;
    case SlZDataBaseItemInfo::ULONG:
        val = INT2FIX(p_zdb->readUlongField(field, card_id));
        break;
    case SlZDataBaseItemInfo::TIME:
        val = rb_funcall(self, rb_intern("read_time_field"), 2,
                         INT2FIX(field), ULONG2NUM(card_id));
        break;
    case SlZDataBaseItemInfo::CATEGORY:
        val = rb_funcall(self, rb_intern("read_categories"), 1,
                         ULONG2NUM(card_id));
        break;
    case SlZDataBaseItemInfo::BINARY:
    case SlZDataBaseItemInfo::COLOR:
    default:
        val = Qnil;
//        rb_raise(rb_eRuntimeError, "Item value_type %d not yet supported.", item_type);
    }
    return val;
}

/*
 * Document-method: write_time_field
 * call-seq:
 *   database.write_time_field(field, val) -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Write the date/time value specified by the field (i.e., index of
 * item_attrs) of the current card.
 */
static VALUE db_write_time_field(VALUE self, VALUE v_field, VALUE val)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    int field = NUM2INT(v_field);

    VALUE result = Qfalse;

    if (RTEST(rb_funcall(val, rb_intern("respond_to?"), 1,
                         rb_str_new2("hour")))) {
        val = rb_funcall(val, rb_intern("new_offset"), 0); // set UTC

        QDate q_date(NUM2INT(rb_funcall(val, rb_intern("year"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("month"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("day"), 0)));
        QTime q_time(NUM2INT(rb_funcall(val, rb_intern("hour"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("min"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("sec"), 0)));

        QDateTime q_datetime(q_date, q_time);
        result = p_zdb->writeField(field, q_datetime)
            ? Qtrue : Qfalse;
    }
    else {
        QDate q_date(NUM2INT(rb_funcall(val, rb_intern("year"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("month"), 0)),
                     NUM2INT(rb_funcall(val, rb_intern("day"), 0)));
        result = p_zdb->writeField(field, q_date)
            ? Qtrue : Qfalse;
    }
    return result;
}        

/*
 * Document-method: write_item
 * call-seq:
 *   database.write_item(key, val) -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Write the value specified by the key of the current card.
 */
static VALUE db_write_item(VALUE self, VALUE key, VALUE val)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    VALUE item_attr_with_index = db_get_item_attr_with_index(self, key);
    if (NIL_P(item_attr_with_index))
        return Qnil;

    VALUE item_attr_index = rb_funcall(item_attr_with_index,
                                       rb_intern("[]"), 1,
                                       INT2FIX(0));

    int field = NUM2INT(item_attr_index);

    VALUE item_attr = rb_funcall(item_attr_with_index,
                                 rb_intern("[]"), 1,
                                 INT2FIX(1));

    int value_type = NUM2INT(rb_funcall(item_attr, rb_intern("value_type"), 0));
    VALUE result = Qfalse;

    switch (value_type) {
    case SlZDataBaseItemInfo::TEXT:
        {
            const char* c_val = StringValuePtr(val);
            result = p_zdb->writeField(field, QString::fromUtf8(c_val))
                ? Qtrue : Qfalse;
        }
        break;
    case SlZDataBaseItemInfo::UCHAR:
        result = p_zdb->writeField(field, (uchar)NUM2INT(val))
            ? Qtrue : Qfalse;
        break;
    case SlZDataBaseItemInfo::USHORT:
        result = p_zdb->writeField(field, (ushort)NUM2INT(val))
            ? Qtrue : Qfalse;
        break;
    case SlZDataBaseItemInfo::ULONG:
        result = p_zdb->writeField(field, (ulong)NUM2ULONG(val))
            ? Qtrue : Qfalse;
        break;
    case SlZDataBaseItemInfo::TIME:
        result = rb_funcall(self, rb_intern("write_time_field"),
                            2, INT2FIX(field), val)
            ? Qtrue : Qfalse;
        break;
    case SlZDataBaseItemInfo::CATEGORY: // must be done afterwards
        result = Qtrue;
    case SlZDataBaseItemInfo::BINARY:
    case SlZDataBaseItemInfo::COLOR:
    default:
        ;
//      p_zdb->cancelEditCard();
//      rb_raise(rb_eRuntimeError, "Item value_type %d not yet supported.", type);

    }
    return result;
}

/*
 * Document-method: update_categories
 * call-seq:
 *   database.upcate_categories(card_id, categories) -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Update the categories off the current card.
 * If any one of the categories does not exist in the CategorySet,
 * add it.
 */
static VALUE db_update_categories(VALUE self, VALUE v_card_id, VALUE categories)
{
    VALUE result = Qfalse;

    VALUE category_set = rb_iv_get(self, "@category_set");
    if (NIL_P(category_set)) {
        category_set = rb_funcall(cCategorySet, rb_intern("new"), 0);
        rb_iv_set(self, "@category_set", category_set);
    }

    VALUE ids = rb_funcall(category_set, rb_intern("get_ids_of_labels"),
                           1, categories);

    QArray<int> q_ids(NUM2INT(rb_funcall(ids, rb_intern("size"), 0)));
    for (QArray<int>::Iterator it = q_ids.begin(); it != q_ids.end(); ++it) {
        *it = NUM2INT(rb_ary_shift(ids));
    }

    SlZDataBase* p_zdb = db_get_zdb(self);

    CardId card_id = NUM2ULONG(v_card_id);
    
    result = p_zdb->updateCategories(card_id, q_ids) ? Qtrue : Qfalse;

    return result;
}

/*
 * Document-method: start_edit_card
 * call-seq:
 *   database.start_edit_card(card_id) -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Start editing the card. You should call this before writing
 * items.
 */
static VALUE db_start_edit_card(VALUE self, VALUE v_card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return p_zdb->startEditCard(NUM2ULONG(v_card_id)) ? Qtrue : Qfalse;
}

/*
 * Document-method: finish_edit_card
 * call-seq:
 *   database.finish_edit_card(card_id) -> fixnum or false
 *
 * (Normally not used directly. Called from Card.)
 * Finish editing the card. You should call this after writing
 * items.
 * Return the new card_id.
 */
static VALUE db_finish_edit_card(VALUE self, VALUE v_card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);
    CardId card_id(NUM2ULONG(v_card_id));

    bool result = p_zdb->finishEditCard(&card_id);

    if (result)
        return ULONG2NUM(card_id);
    else
        return Qfalse;
}

/*
 * Document-method: cancel_edit_card
 * call-seq:
 *   database.cancel_edit_card -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Cancel editing the card.
 */
static VALUE db_cancel_edit_card(VALUE self)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    return p_zdb->cancelEditCard() ? Qtrue : Qfalse;
}

/*
 * Document-method: delete_card
 * call-seq:
 *   database.delete_card -> true or false
 *
 * (Normally not used directly. Called from Card.)
 * Delete the card.
 */
static VALUE db_delete_card(VALUE self, VALUE v_card_id)
{
    SlZDataBase* p_zdb = db_get_zdb(self);

    CardId card_id(NUM2ULONG(v_card_id));

    return p_zdb->deleteCard(&card_id) ? Qtrue : Qfalse;
}

/* for debug */
static VALUE db_inspect_fields(VALUE self)
{
    VALUE result = rb_ary_new();

    SlZDataBase* p_zdb = db_get_zdb(self);

    int count(p_zdb->items().count());

    VALUE used_field_ids = rb_ary_new();
    int field_id;

    for (int i = 0; i < count; ++i) {
        VALUE info = rb_ary_new();

                 field_id          = p_zdb->nameToFieldId(i);
        QString  item_name         ( p_zdb->itemName(i));
        DataType field_type        = p_zdb->readFieldType(i);
        QChar    another_item_type ( item_name.at(4));

        rb_ary_push(info, INT2FIX(field_id));
        rb_ary_push(info, rb_str_new2(item_name.left(4).latin1()));
        rb_ary_push(info, INT2FIX(field_type));
        rb_ary_push(info, INT2FIX(another_item_type.latin1()));

        rb_ary_push(used_field_ids, INT2FIX(field_id));

        rb_ary_push(result, info);
    }

    return result;
}

static VALUE db_find_item_info(VALUE self, VALUE name)
{
    VALUE result = Qnil;

    SlZDataBase* p_zdb = db_get_zdb(self);

    QString q_name(QString::fromUtf8(StringValuePtr(name)));

    int item_id = p_zdb->nameToItemId(q_name);

    if (item_id > 0) {
        result = rb_ary_new();

        int      item_id   = p_zdb->nameToItemId(q_name);
        DataType item_type = p_zdb->readItemType(q_name);

        rb_ary_push(result, name);
        rb_ary_push(result, INT2FIX(item_id));
        rb_ary_push(result, INT2FIX(item_type));
    }

    return result;
}


static const char* ITEMATTR_TYPES[] = {
    "TYPE_TEXT",
    "TYPE_UCHAR",
    "TYPE_USHORT",
    "TYPE_ULONG",
    "TYPE_COLOR",
    "TYPE_TIME",
    "TYPE_BINARY",
    "TYPE_CATEGORY",
};

/*
 * Document-method: initialize
 * call-seq:
 *   ItemAttr.new(name, value_type) -> itemattr
 *
 * Create a new ItemAttr object.
 * Possible value_type values are the following:
 *
 *    ItemAttr::TYPE_TEXT
 *    ItemAttr::TYPE_UCHAR
 *    ItemAttr::TYPE_USHORT
 *    ItemAttr::TYPE_ULONG
 *    ItemAttr::TYPE_COLOR
 *    ItemAttr::TYPE_TIME
 *    ItemAttr::TYPE_BINARY
 *    ItemAttr::TYPE_CATEGORY
 */
static VALUE ia_initialize(VALUE self, VALUE name, VALUE value_type)
{
    rb_iv_set(self, "@name", name);
    rb_iv_set(self, "@value_type", value_type);
    return self;
}




/*
 * Document-method: initialize
 * call-seq:
 *   Card.new(database, card_id) -> card
 *
 * Create a new Card object. You should specify valid DataBase objec
 * and card_id (Integer).
 * If card_id is 0, it is treated as a new card.
 */
static VALUE card_initialize(VALUE self, VALUE database, VALUE card_id)
{
    rb_iv_set(self, "@database", database);
    rb_iv_set(self, "@card_id", card_id);

    VALUE values_cache = rb_hash_new();
    rb_iv_set(self, "values_cache", values_cache);

    return self;
}

static VALUE each(VALUE obj)
{
    return rb_funcall(obj, rb_intern("each"), 0);
}


static VALUE card_append_key(VALUE item_attr, VALUE self)
{
    VALUE keys = rb_iv_get(self, "keys");
    VALUE key = rb_iv_get(item_attr, "@name");
    rb_ary_push(keys, key);
    return key;
}

/*
 * Document-method: keys
 * call-seq:
 *   card.keys -> array
 *
 * Return keys of the card.
 */
static VALUE card_keys(VALUE self)
{
    VALUE database = rb_iv_get(self, "@database");
    VALUE keys = rb_iv_get(self, "keys");
    if (keys != Qnil)
	return keys;

    keys = rb_ary_new();
    rb_iv_set(self, "keys", keys);

    rb_iterate(RUBY_METHOD_FUNC(each),
	       rb_iv_get(database, "@item_attrs"),
	       RUBY_METHOD_FUNC(card_append_key),
	       self);

    return keys;
}

/*
 * Document-method: []
 * call-seq:
 *   card[key] -> object
 *
 * Return the item associated with key.
 * The type of the item is determined by the corresponding element
 * of database.item_attrs.
 */
static VALUE card_read_item(VALUE self, VALUE key)
{
    VALUE val = Qnil;

    VALUE values_cache = rb_iv_get(self, "values_cache");

    val = rb_funcall(values_cache, rb_intern("[]"), 1, key);

    if (!NIL_P(val)) {
        return val;
    }

    VALUE db = rb_iv_get(self, "@database");

    val = rb_funcall(db, rb_intern("read_item"), 2,
                      key, rb_iv_get(self, "@card_id"));

    rb_funcall(values_cache, rb_intern("[]="), 2, key, val);

    return val;
}

/*
 * Document-method: []=
 * call-seq:
 *   card[key] = val
 *
 * Assign the value associated with the key of the card.
 * Values are stored temporary on memory.
 * To save data, call card.save.
 * The type of the value should correspond to the element
 * of database.item_attrs.
 */

static VALUE card_write_item(VALUE self, VALUE key, VALUE val)
{
    VALUE values_cache = rb_iv_get(self, "values_cache");
    rb_funcall(values_cache, rb_intern("[]="), 2, key, val);
    return val;
}

VALUE card_save_i(VALUE key, VALUE self)
{
    VALUE db = rb_iv_get(self, "@database");
    VALUE values_cache = rb_iv_get(self, "values_cache");
    VALUE val = rb_funcall(values_cache, rb_intern("[]"),1, key);

    if (NIL_P(val))
        return Qnil;

    VALUE result = rb_funcall(db, rb_intern("write_item"), 2, key, val);
    if (!RTEST(result)) {
        rb_funcall(db, rb_intern("cancel_edit_card"), 0);
        rb_raise(rb_eRuntimeError, "Save item %s failed", StringValuePtr(key));
    }
    return result;
}

/*
 * Document-method: save
 * call-seq:
 *   card.save -> true or nil
 *
 * Save the temporary data to database file.
 */

static VALUE card_save(VALUE self)
{
    VALUE db           = rb_iv_get(self, "@database");
    VALUE card_id      = rb_iv_get(self, "@card_id");

    if (!RTEST(rb_funcall(db, rb_intern("start_edit_card"), 1, card_id))) {
        rb_raise(rb_eRuntimeError,
                 "Start editing card %d failed.", NUM2ULONG(card_id));
        return Qnil;
    }                         

    rb_iterate(RUBY_METHOD_FUNC(each),
               rb_funcall(self, rb_intern("keys"), 0),
               RUBY_METHOD_FUNC(card_save_i),
               self);
    

    card_id = rb_funcall(db, rb_intern("finish_edit_card"), 1, card_id);
    if (!RTEST(card_id)) {
        rb_raise(rb_eRuntimeError,
                 "Finish editing card %d failed.", NUM2ULONG(card_id));
        return Qnil;
    }

    rb_iv_set(self, "@card_id", card_id);

    VALUE categories = rb_funcall(self, rb_intern("categories"), 0);
    if (!NIL_P(categories)
        && NUM2INT(rb_funcall(categories, rb_intern("size"), 0)) > 0) {

        VALUE result = rb_funcall(db, rb_intern("update_categories"), 2,
                                  categories, card_id);
        if (!RTEST(result)) {
            rb_raise(rb_eRuntimeError,
                     "Updating categories of %d failed.", NUM2ULONG(card_id));
            return Qnil;
        }
    }

    return Qtrue;
}

/*
 * Document-method: delete
 * call-seq:
 *   card.delete -> true or false
 *
 * Delete this card.
 *
 */
static VALUE card_delete(VALUE self)
{
    VALUE card_id = rb_iv_get(self, "@card_id");
    VALUE database = rb_iv_get(self, "@database");

    return rb_funcall(database, rb_intern("delete_card"), 1, card_id);
}

static VALUE card_yield_value(VALUE item_attr, VALUE card)
{
    VALUE key = rb_iv_get(item_attr, "name");
    rb_yield(rb_funcall(card, rb_intern("[]"), 1, key));
    return Qnil;
}

/*
 * Document-method: each
 * call-seq:
 *   card.each {|val| ... } -> card
 *
 * Iterate the values of the card. The order is the same as of the
 * item_attrs of the database.
 */
static VALUE card_each(VALUE self)
{
    VALUE db = rb_iv_get(self, "@database");
    VALUE item_attrs = rb_iv_get(db, "@item_attrs");

    rb_iterate(RUBY_METHOD_FUNC(each), item_attrs,
	       RUBY_METHOD_FUNC(card_yield_value), self);

    return self;
}

static VALUE card_add_item_to_result_hash(VALUE item_attr, VALUE card)
{
    VALUE key = rb_iv_get(item_attr, "@name");
    VALUE val = rb_funcall(card, rb_intern("[]"), 1, key);
    VALUE result = rb_iv_get(card, "tmp_result");
    rb_funcall(result, rb_intern("[]="), 2, key, val);
    return Qnil;
}

/*
 * Document-method: to_hash
 * call-seq:
 *   card.to_hash -> hash
 *
 * Return key-value pairs of the card as hash.
 */
static VALUE card_to_hash(VALUE self)
{
    VALUE db = rb_iv_get(self, "@database");
    VALUE item_attrs = rb_iv_get(db, "@item_attrs");

    VALUE result = rb_hash_new();

    rb_iv_set(self, "tmp_result", result);

    rb_iterate(RUBY_METHOD_FUNC(each), item_attrs,
	       RUBY_METHOD_FUNC(card_add_item_to_result_hash), self);

    return result;
}

/* implementation depended */

/*
 * Document-method: categories
 * call-seq:
 *   card.categories -> array
 *
 * Return categories of the card.
 */
static VALUE card_get_categories(VALUE self)
{
    return rb_funcall(self, rb_intern("[]"), 1, rb_str_new2("NONAME0"));
}

/*
 * Document-method: categories=
 * call-seq:
 *   card.categories = ["string", "string", ...]
 *
 * Assign the categories of the card.
 */
static VALUE card_set_categories(VALUE self, VALUE new_cats)
{
    return rb_funcall(self, rb_intern("[]="), 2,
                      rb_str_new2("NONAME0"),
                      new_cats);
}


static void cs_free(void* p)
{
    delete static_cast<SlCategories*>(p);
}

static VALUE cs_alloc(VALUE klass)
{
    SlCategories* p_categories;
    VALUE obj;

    p_categories = new SlCategories();
    obj = Data_Wrap_Struct(klass, 0, cs_free, p_categories);
    return obj;
}

SlCategories* cs_get_slcats(VALUE self)
{
    SlCategories* p_slcats;
    Data_Get_Struct(self, SlCategories, p_slcats);
    return p_slcats;
}
/*
 * Document-method: []
 * call-seq:
 *   category_set[id] -> string
 *
 * Return the category label specified by id.
 */
static VALUE cs_get_label(VALUE self, VALUE id)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(p_slcats->label(NUM2INT(id)));

    return rb_str_new2(q_label.utf8());
}

/*
 * Document-method: id_of
 * call-seq:
 *   category_set.id_of(label) -> fixnum or nil.
 *
 * Return the id of the label.
 * Return nil if the label does not exist in the category set.
 */
static VALUE cs_get_id_of_label(VALUE self, VALUE label)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(QString::fromUtf8(StringValuePtr(label)));

    int id = p_slcats->id(q_label);

    return id ? INT2FIX(id) : Qnil;
}

/*
 * Document-method: labels
 * call-seq:
 *   category_set.labels -> array
 */
static VALUE cs_labels(VALUE self)
{
    VALUE result = rb_ary_new();

    SlCategories* p_slcats = cs_get_slcats(self);

    QStringList q_labels(p_slcats->labels());
    for (QStringList::ConstIterator it = q_labels.begin();
	 it != q_labels.end();
	 ++it) {
	QString q_label(*it);
	VALUE label = rb_str_new2(q_label.utf8());
	rb_ary_push(result, label);
    }
    return result;
}

static VALUE cs_each_i(VALUE label, VALUE self)
{
    return rb_yield(label);
}

/*
 * Document-method: each
 * call-seq:
 *   category_set.each {|label| ... } -> category_set
 */
static VALUE cs_each(VALUE self)
{
    rb_iterate(RUBY_METHOD_FUNC(each),
               rb_funcall(self, rb_intern("labels"), 0),
	       RUBY_METHOD_FUNC(cs_each_i),
               self);
	       
    return self;
}
 
/*
 * Document-method: include?
 * call-seq:
 *   category_set.include?(label) -> true or false
 *
 * Return true if the label exist in the category set.
 */
static VALUE cs_include_p(VALUE self, VALUE label)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(QString::fromUtf8(StringValuePtr(label)));

    return p_slcats->exists(q_label) ? Qtrue : Qfalse;
}

/*
 * Document-method: add
 * call-seq:
 *   category_set.add(label) -> fixnum
 *
 * Add the label to the category set. Return the id of the category.
 */
static VALUE cs_add(VALUE self, VALUE label)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(QString::fromUtf8(StringValuePtr(label)));

    int new_id = p_slcats->addCategory(q_label);

    return INT2FIX(new_id);
}

/*
 * Document-method: delete
 * call-seq:
 *   category_set.delete(label) -> true or false
 *
 */
static VALUE cs_delete(VALUE self, VALUE label)
{
    SlCategories* p_slcats = cs_get_slcats(self);

    QString q_label(QString::fromUtf8(StringValuePtr(label)));

    bool result = p_slcats->removeCategory(q_label);

    return result ? Qtrue : Qfalse;
}

VALUE cs_get_ids_of_labels_i(VALUE label, VALUE args)
{
    VALUE result = rb_funcall(args, rb_intern("[]"), 1, INT2FIX(0));
    VALUE category_set = rb_funcall(args, rb_intern("[]"), 1, INT2FIX(1));

    if (!RTEST(rb_funcall(category_set, rb_intern("include?"), 1,
                          label))) {
        rb_funcall(category_set, rb_intern("add"), 1, label);
    }

    VALUE id = rb_funcall(category_set, rb_intern("id_of"), 1, label);

    rb_ary_push(result, id);
    return Qnil;
}

/*
 * Document-method: get_ids_of_labels
 * call-seq:
 *   category_set.get_ids_of_labels(labels) -> array
 */
static VALUE cs_get_ids_of_labels(VALUE self, VALUE labels)
{
    VALUE result = rb_ary_new();
    VALUE args = rb_ary_new();
    rb_ary_push(args, result);
    rb_ary_push(args, self);

    rb_iterate(RUBY_METHOD_FUNC(each),
               labels,
               RUBY_METHOD_FUNC(cs_get_ids_of_labels_i),
               args);
    return result;
}


extern "C" void Init_simplezdb()
{
    rb_require("date");
    cDate     = rb_funcall(rb_cObject, rb_intern("const_get"), 1,
                           rb_str_new2("Date"));
    cDateTime = rb_funcall(rb_cObject, rb_intern("const_get"), 1,
                           rb_str_new2("DateTime"));

    mSimpleZDB = rb_define_module("SimpleZDB");

    cDataBase = rb_define_class_under(mSimpleZDB,
				      "DataBase",
				      rb_cObject);
    rb_include_module(cDataBase, rb_mEnumerable);
    rb_define_attr(cDataBase,   "item_attrs", 1, 0);

    rb_define_singleton_method(cDataBase, "open", RUBY_METHOD_FUNC(db_s_open), -1);

    rb_define_method(cDataBase, "initialize", RUBY_METHOD_FUNC(db_initialize), -2);
    rb_define_method(cDataBase, "close",      RUBY_METHOD_FUNC(db_close), 0);

    rb_define_method(cDataBase, "size",       RUBY_METHOD_FUNC(db_size), 0);
    rb_define_alias(cDataBase, "length", "size");
    rb_define_method(cDataBase, "each",       RUBY_METHOD_FUNC(db_each), 0);


    rb_define_method(cDataBase, "read_time_field", RUBY_METHOD_FUNC(db_read_time_field), 2);
    rb_define_method(cDataBase, "read_categories", RUBY_METHOD_FUNC(db_read_categories), 1);
    rb_define_method(cDataBase, "read_item", RUBY_METHOD_FUNC(db_read_item), 2);
    rb_define_method(cDataBase, "write_time_field", RUBY_METHOD_FUNC(db_write_time_field), 2);
    rb_define_method(cDataBase, "write_item", RUBY_METHOD_FUNC(db_write_item), 2);
    rb_define_method(cDataBase, "update_categories", RUBY_METHOD_FUNC(db_update_categories), 2);
    rb_define_method(cDataBase, "start_edit_card", RUBY_METHOD_FUNC(db_start_edit_card), 1);
    rb_define_method(cDataBase, "finish_edit_card", RUBY_METHOD_FUNC(db_finish_edit_card), 1);
    rb_define_method(cDataBase, "cancel_edit_card", RUBY_METHOD_FUNC(db_cancel_edit_card), 0);
    rb_define_method(cDataBase, "delete_card",      RUBY_METHOD_FUNC(db_delete_card), 1);
    rb_define_method(cDataBase, "inspect_fields", RUBY_METHOD_FUNC(db_inspect_fields), 0);
    rb_define_method(cDataBase, "find_item_info", RUBY_METHOD_FUNC(db_find_item_info), 1);

    size_t db_types_size(sizeof(DATABASE_TYPES)/sizeof(char*));
    for (size_t i = 0; i < db_types_size; ++i)
        rb_iv_set(cDataBase, DATABASE_TYPES[i], INT2FIX(i));

    cItemAttr = rb_define_class_under(mSimpleZDB,
				      "ItemAttr",
				      rb_cObject);
    rb_define_attr(cItemAttr, "name", 1, 0);
    rb_define_attr(cItemAttr, "value_type", 1, 0);
    rb_define_method(cItemAttr, "initialize", RUBY_METHOD_FUNC(ia_initialize), 2);

    size_t ia_types_size(sizeof(ITEMATTR_TYPES)/sizeof(char*));
    for (size_t i = 0; i < ia_types_size; ++i)
        rb_iv_set(cItemAttr, ITEMATTR_TYPES[i], INT2FIX(i));

    cCard = rb_define_class_under(mSimpleZDB,
				   "Card",
				   rb_cObject);
    rb_include_module(cCard, rb_mEnumerable);
    rb_define_attr(cCard,   "card_id",     1, 0);
    rb_define_attr(cCard,   "database",    1, 0);
    rb_define_method(cCard, "initialize",  RUBY_METHOD_FUNC(card_initialize), 2);
    rb_define_method(cCard, "keys",        RUBY_METHOD_FUNC(card_keys), 0);
    rb_define_method(cCard, "[]",          RUBY_METHOD_FUNC(card_read_item), 1);
    rb_define_method(cCard, "[]=",         RUBY_METHOD_FUNC(card_write_item), 2);
    rb_define_method(cCard, "save",        RUBY_METHOD_FUNC(card_save), 0);
    rb_define_method(cCard, "delete",      RUBY_METHOD_FUNC(card_delete), 0);
    rb_define_method(cCard, "each",        RUBY_METHOD_FUNC(card_each), 0);
    rb_define_alias(cCard, "to_ary", "to_a");
    rb_define_method(cCard, "to_hash",     RUBY_METHOD_FUNC(card_to_hash), 0);
    rb_define_method(cCard, "categories",  RUBY_METHOD_FUNC(card_get_categories), 0);
    rb_define_method(cCard, "categories=", RUBY_METHOD_FUNC(card_set_categories), 1);

    cCategorySet = rb_define_class_under(mSimpleZDB,
					 "CategorySet",
					 rb_cObject);

    rb_include_module(cCategorySet, rb_mEnumerable);

    rb_define_alloc_func(cCategorySet, RUBY_METHOD_FUNC(cs_alloc));
    rb_define_method(cCategorySet, "[]",       RUBY_METHOD_FUNC(cs_get_label), 1);
    rb_define_method(cCategorySet, "id_of",    RUBY_METHOD_FUNC(cs_get_id_of_label), 1);
    rb_define_method(cCategorySet, "labels",   RUBY_METHOD_FUNC(cs_labels), 0);
    rb_define_method(cCategorySet, "each",     RUBY_METHOD_FUNC(cs_each), 0);
    rb_define_alias(cCategorySet, "to_ary", "to_a");
    rb_define_method(cCategorySet, "include?", RUBY_METHOD_FUNC(cs_include_p), 1);
    rb_define_method(cCategorySet, "add",      RUBY_METHOD_FUNC(cs_add), 1);
    rb_define_method(cCategorySet, "delete",   RUBY_METHOD_FUNC(cs_delete), 1);
    rb_define_method(cCategorySet, "get_ids_of_labels", RUBY_METHOD_FUNC(cs_get_ids_of_labels), 1);
}
